#include <stdio.h>
#include <stdarg.h>
#include <asm/unistd.h>
#include <errno.h>
#include "sysmp.h"

#define __NR__sysmp 	__NR_sysmp

static inline _syscall5(int, _sysmp, int, cmd, 
	int, i1, int, i2, int, i3, unsigned long, ul)

/* this is the entry point for user applications/libs */
int sysmp(int cmd, ...) 
{
	va_list		va;
	int		i1, i2, i3;
	unsigned long	ul;
	int		result;

	va_start(va, cmd);

	switch (cmd) {
		/* these have no extra parameters */
		case MP_PGSIZE:
		case MP_SCHED:
		case MP_NPROCS:
		case MP_NAPROCS:
		case MP_CURPROC:
		case MP_GETMUSTRUN:
		case MP_RUNANYWHERE:
			result = _sysmp(cmd, 0, 0, 0, (long)0);
			break;

		/* these have 1 int parameter */
		case MP_RESTRICT:
		case MP_EMPOWER:
		case MP_ISOLATE:
		case MP_UNISOLATE:
		case MP_DISABLE:
		case MP_ENABLE:
		case MP_NONPREEMPTIVE:
		case MP_PREEMPTIVE:
		case MP_CLOCK:
		case MP_FASTCLOCK:
		case MP_MUSTRUN:
		case MP_GETMUSTRUN_PID:
		case MP_RUNANYWHERE_PID:
		case MP_KERNADDR:
		case MP_SASZ:
			i1 = va_arg(va, int);
			result = _sysmp(cmd, i1, 0, 0, (long)0);
			break;

		/* this uses pda_stat * - use ul, assume same length */
		case MP_STAT:
			ul = va_arg(va, unsigned long);
			result = _sysmp(cmd, 0, 0, 0, ul);
			break;

		/* this uses int and pid_t */
		case MP_MUSTRUN_PID:
			i1 = va_arg(va, int);
			i2 = va_arg(va, int);
			result = _sysmp(cmd, i1, i2, 0, (long)0);
			break;

		/* this uses an int, char * and int */
		case MP_SAGET:
			i1 = va_arg(va, int);
			ul = va_arg(va, unsigned long);
			i2 = va_arg(va, int);
			result = _sysmp(cmd, i1, i2, 0, ul);
			break;

		/* this uses 3 ints and a char * */
		case MP_SAGET1:
			i1 = va_arg(va, int);
			ul = va_arg(va, unsigned long);
			i2 = va_arg(va, int);
			i3 = va_arg(va, int);
			result = _sysmp(cmd, i1, i2, i3, ul);
			break;

		/* This uses two ints */
		case MP_PSET:
			i1 = va_arg(va, int);
			i2 = va_arg(va, int);
			switch(i1) {
				case MPPS_DELETE:
				  result = _sysmp(cmd, i1, i2, 0, (long)0);
				  break;
				case MPPS_CREATE:
				case MPPS_ADD:
				case MPPS_REMOVE:
				  ul = va_arg(va, cpumask_t);
				  result = _sysmp(cmd, i1, i2, 0, ul);
				  break;
				default:
					result = -EINVAL;
					break;
			}
			break;	
		default:
			result = -EINVAL;
			break;	
	}
	va_end(va);

	return result;
}
