/*
 * mpadmin
 * control and report processor status
 *
 * copyright 1998-1999 Tim Hockin
 * Released under the GNU Public License (GPL)
 *
 * Known bugs: none
 * Added from SGI specs: -e -d -p
 * Not implemented: -c -f -D -C
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "sysmp.h"

void version()
{
	fprintf(stderr,"mpadmin - version %s by Tim Hockin\n",PSET_UTILS_VER);
}

void usage()
{
	version();
	fprintf(stderr, 
	"Usage: mpadmin  -s | -n | -v | -r [cpu] | -u [cpu] | -d [cpu] |\n"); 
	fprintf(stderr, 
	"                -e [cpu] | -I [cpu] | -U [cpu] | -p [cpu] |\n"); 
	fprintf(stderr, 
	"                -c [cpu] | -f [cpu] | -D [cpu] | -C [cpu]\n");
	fprintf(stderr,
	"\n        options:\n");
	fprintf(stderr,
	"          -s              summary of settings\n");
	fprintf(stderr, 
	"          -n              list of all processors\n");
	fprintf(stderr,
	"          -v              version info\n");
	fprintf(stderr,
	"          -r [cpu]        list of restricted CPUs or restrict CPU [cpu]\n");
	fprintf(stderr,
	"          -u [cpu]        list of empowered CPUs or empower CPU [cpu]\n");
	fprintf(stderr,  
	"          -d [cpu]        list of disabled CPUs or disable CPU [cpu]\n");
	fprintf(stderr,
	"          -e [cpu]        list of enabled CPUs or enable CPU [cpu]\n");
	fprintf(stderr,
	"          -I [cpu]        list of isolated CPUs or isolate CPU [cpu]\n"); 
	fprintf(stderr,
	"          -U [cpu]        list of unisolated CPUs or unisolate CPU [cpu]\n");
	fprintf(stderr,
	"          -p [cpu]        list of preemptive CPUs or mark as preemptive CPU [cpu]\n");
	fprintf(stderr,
	"          -c              for compatibilty - function not supported\n");
	fprintf(stderr,
	"          -f              for compatibilty - function not supported\n");
	fprintf(stderr,
	"          -D              for compatibilty - function not supported\n");
	fprintf(stderr,
	"          -C              for compatibilty - function not supported\n");
 }

int intparam(char *p)
{
	if (!isdigit(*p))
		return -1;

	return atoi(p);
}

/* handle the n option - a list of all processors*/
void opt_n(void)
{
	struct pda_stat *array;
	int nprocs, i=0;

	nprocs = sysmp(MP_NPROCS);
	array = (struct pda_stat *)malloc(sizeof(struct pda_stat)*nprocs);
	if (array == NULL) {
		fprintf(stderr, "Can not allocate resources.\n");
		return;
	}
	sysmp(MP_STAT, array);
	for (i = 0; i < nprocs; i++) {
		printf("%d\n", array[i].p_cpuid);
	}
	
	free(array);
	return;
}

/* handle the s option - a summary */
void opt_s(void)
{
	struct pda_stat *array;
	int nprocs, i=0;

	nprocs = sysmp(MP_NPROCS);
	array = (struct pda_stat *)malloc(sizeof(struct pda_stat)*nprocs);
	if (array == NULL) {
		fprintf(stderr, "Can not allocate resources.\n");
		return;
	}

	sysmp(MP_STAT, array);

	printf("processors: ");
	for (i = 0; i < nprocs; i++) {
		printf("%d ", array[i].p_cpuid);
	}
	printf("\n");
		
	printf("enabled: ");
	for (i = 0; i < nprocs; i++) {
		if (array[i].p_flags & CPUF_ENABLED)
			printf("%d ", array[i].p_cpuid);
	}
	printf("\n");

	printf("restricted: ");
	for (i = 0; i < nprocs; i++) {
		if (array[i].p_flags & CPUF_RESTRICTED)
			printf("%d ", array[i].p_cpuid);
	}
	printf("\n");

	printf("isolated: ");
	for (i = 0; i < nprocs; i++) {
		if (array[i].p_flags & CPUF_ISOLATED)
			printf("%d ", array[i].p_cpuid);
	}
	printf("\n");

	printf("disabled: ");
	for (i = 0; i < nprocs; i++) {
		if (array[i].p_flags & CPUF_DISABLED)
			printf("%d ", array[i].p_cpuid);
	}
	printf("\n");

	printf("non-preemptive: ");
	for (i = 0; i < nprocs; i++) {
		if (array[i].p_flags & CPUF_NONPREEMPTIVE)
			printf("%d ", array[i].p_cpuid);
	}
	printf("\n");

	free(array);
	
	return;
}

/* empower a CPU, or print empowered CPU's */
int opt_u(int cpu)
{
	struct pda_stat *array;
	int nprocs, i=0;
	int result = 0;

	nprocs = sysmp(MP_NPROCS);
	array = (struct pda_stat *)malloc(sizeof(struct pda_stat)*nprocs);
	if (array == NULL) {
		perror("malloc");
		return -1;
	}
	sysmp(MP_STAT, array);
	
	/* print empowered CPUs */
	if (cpu < 0) {
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_flags == CPUF_ENABLED)
				printf("%d\n", array[i].p_cpuid);
		}
	} else {
		/* find the right CPU */
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_cpuid == cpu) {
				result = sysmp(MP_EMPOWER, cpu);
				if (result)
					perror("sysmp(EMPOWER)");
				break;
			}
		}
		/* requested CPU does not exist */
		if (i == nprocs) {
			result = -1;
			fprintf(stderr, "CPU %d does not exist\n", cpu);
		}
	}

	free(array);
	return result;
}

/* restrict a CPU, or print restricted CPU's */
int opt_r(int cpu)
{
	struct pda_stat *array;
	int nprocs, i=0;
	int result = 0;

	nprocs = sysmp(MP_NPROCS);
	array = (struct pda_stat *)malloc(sizeof(struct pda_stat)*nprocs);
	if (array == NULL) {
		perror("malloc");
		return -1;
	}
	sysmp(MP_STAT, array);
	
	/* print restricted CPUs */
	if (cpu < 0) {
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_flags == CPUF_RESTRICTED)
				printf("%d\n", array[i].p_cpuid);
		}
	} else {
		/* find the right CPU */
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_cpuid == cpu) {
				result = sysmp(MP_RESTRICT, cpu);
				if (result)
					perror("sysmp(RESTRICT)");
				break;
			}
		}
		/* requested CPU does not exist */
		if (i == nprocs) {
			result = -1;
			fprintf(stderr, "CPU %d does not exist\n", cpu);
		}
	}

	free(array);
	return result;
}

/* enable a CPU, or print enabled CPU's */
int opt_e(int cpu)
{
	struct pda_stat *array;
	int nprocs, i=0;
	int result = 0;

	nprocs = sysmp(MP_NPROCS);
	array = (struct pda_stat *)malloc(sizeof(struct pda_stat)*nprocs);
	if (array == NULL) {
		perror("malloc");
		return -1;
	}
	sysmp(MP_STAT, array);
	
	/* print enabled CPUs */
	if (cpu < 0) {
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_flags == CPUF_ENABLED)
				printf("%d\n", array[i].p_cpuid);
		}
	} else {
		/* find the right CPU */
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_cpuid == cpu) {
				result = sysmp(MP_ENABLE, cpu);
				if (result)
					perror("sysmp(ENABLE)");
				break;
			}
		}
		/* requested CPU does not exist */
		if (i == nprocs) {
			result = -1;
			fprintf(stderr, "CPU %d does not exist\n", cpu);
		}
	}

	free(array);
	return result;
}

/* disable a CPU, or print disabled CPU's */
int opt_d(int cpu)
{
	struct pda_stat *array;
	int nprocs, i=0;
	int result = 0;

	nprocs = sysmp(MP_NPROCS);
	array = (struct pda_stat *)malloc(sizeof(struct pda_stat)*nprocs);
	if (array == NULL) {
		perror("malloc");
		return -1;
	}
	sysmp(MP_STAT, array);
	
	/* print disabled CPUs */
	if (cpu < 0) {
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_flags == CPUF_DISABLED)
				printf("%d\n", array[i].p_cpuid);
		}
	} else {
		/* find the right CPU */
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_cpuid == cpu) {
				result = sysmp(MP_DISABLE, cpu);
				if (result)
					perror("sysmp(DISABLE)");
				break;
			}
		}
		/* requested CPU does not exist */
		if (i == nprocs) {
			result = -1;
			fprintf(stderr, "CPU %d does not exist\n", cpu);
		}
	}

	free(array);
	return result;
}

/* unisolate a CPU, or print unisolated CPU's */
int opt_U(int cpu)
{
	struct pda_stat *array;
	int nprocs, i=0;
	int result = 0;

	nprocs = sysmp(MP_NPROCS);
	array = (struct pda_stat *)malloc(sizeof(struct pda_stat)*nprocs);
	if (array == NULL) {
		perror("malloc");
		return -1;
	}
	sysmp(MP_STAT, array);
	
	/* print unisolated CPUs */
	if (cpu < 0) {
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_flags == CPUF_ENABLED)
				printf("%d\n", array[i].p_cpuid);
		}
	} else {
		/* find the right CPU */
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_cpuid == cpu) {
				result = sysmp(MP_UNISOLATE, cpu);
				if (result)
					perror("sysmp(UNISOLATE)");
				break;
			}
		}
		/* requested CPU does not exist */
		if (i == nprocs) {
			result = -1;
			fprintf(stderr, "CPU %d does not exist\n", cpu);
		}
	}

	free(array);
	return result;
}

/* isolate a CPU, or print isolated CPU's */
int opt_I(int cpu)
{
	struct pda_stat *array;
	int nprocs, i=0;
	int result = 0;

	nprocs = sysmp(MP_NPROCS);
	array = (struct pda_stat *)malloc(sizeof(struct pda_stat)*nprocs);
	if (array == NULL) {
		perror("malloc");
		return -1;
	}
	sysmp(MP_STAT, array);
	
	/* print isolated CPUs */
	if (cpu < 0) {
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_flags == CPUF_ISOLATED)
				printf("%d\n", array[i].p_cpuid);
		}
	} else {
		/* find the right CPU */
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_cpuid == cpu) {
				result = sysmp(MP_ISOLATE, cpu);
				if (result)
					perror("sysmp(ISOLATE)");
				break;
			}
		}
		/* requested CPU does not exist */
		if (i == nprocs) {
			result = -1;
			fprintf(stderr, "CPU %d does not exist\n", cpu);
		}
	}

	free(array);
	return result;
}

/* make a CPU preemptive, or print preemptive CPU's */
int opt_p(int cpu)
{
	struct pda_stat *array;
	int nprocs, i=0;
	int result = 0;

	nprocs = sysmp(MP_NPROCS);
	array = (struct pda_stat *)malloc(sizeof(struct pda_stat)*nprocs);
	if (array == NULL) {
		perror("malloc");
		return -1;
	}
	sysmp(MP_STAT, array);
	
	/* print preemptive CPUs */
	if (cpu < 0) {
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_flags == CPUF_ENABLED)
				printf("%d\n", array[i].p_cpuid);
		}
	} else {
		/* find the right CPU */
		for (i = 0; i < nprocs; i++) {
			if (array[i].p_cpuid == cpu) {
				result = sysmp(MP_PREEMPTIVE, cpu);
				if (result)
					perror("sysmp(PREEMPTIVE)");
				break;
			}
		}
		/* requested CPU does not exist */
		if (i == nprocs) {
			result = -1;
			fprintf(stderr, "CPU %d does not exist\n", cpu);
		}
	}

	free(array);
	return result;
}


/* **********************************************************************
 * Main()
 * *********************************************************************/
int main(int argc, char *argv[])
{
	int retval = 0;

	if (argc < 2 || argc > 3 || argv[1][0] != '-') {
		usage();
		return -1;
	}
	
	switch(argv[1][1]) {
		case 'n':	
			if (argc == 2) {
				opt_n();
			} else {
				usage();
				retval = -1;
			}
			break;
		case 's':	
			if (argc == 2) { 
				opt_s();
			} else { 
				usage();
				retval = -1;
			}
			break;
		case 'u':	
			if (argc == 3) {
				/* we expect something like -u 3 */
				retval = opt_u(intparam(argv[2]));
			} else {
				/* we expect something like -u3 */
				retval = opt_u(intparam(&argv[1][2]));
			}		
			break;
		case 'r':	
			if (argc == 3) {
				retval = opt_r(intparam(argv[2]));
			} else {
				retval = opt_r(intparam(&argv[1][2]));
			}		
			break;
		case 'e':
			if (argc == 3) {
				retval = opt_e(intparam(argv[2]));
			} else {
				retval = opt_e(intparam(&argv[1][2]));
			}		
			break;
		case 'd':	
			if (argc == 3) {
				retval = opt_d(intparam(argv[2]));
			} else {
				retval = opt_d(intparam(&argv[1][2]));
			}		
			break;
		case 'U':	
			if (argc == 3) {
				retval = opt_U(intparam(argv[2]));
			} else {
				retval = opt_U(intparam(&argv[1][2]));
			}		
			break;
		case 'I':
			if (argc == 3) {
				retval = opt_I(intparam(argv[2]));
			} else {
				retval = opt_I(intparam(&argv[1][2]));
			}		
			break;
		case 'p':	
			if (argc == 3) {
				retval = opt_p(intparam(argv[2]));
			} else {
				retval = opt_p(intparam(&argv[1][2]));
			}		
			break;
		case 'c':	
			fprintf(stderr, "function not supported\n");
			break;
		case 'f':	
			fprintf(stderr, "function not supported\n");
			break;
		case 'D':	
			fprintf(stderr, "function not supported\n");
			break;
		case 'C':	
			fprintf(stderr, "function not supported\n");
			break;
		case 'v':	
			version();
			break;
		default: 	
			usage();
			retval = -1;
			break;
	}
	return retval;
}
