/*
 * pset
 * not implemented yet
 *
 * copyright 1998-1999 Tim Hockin
 * Released under the GNU Public License (GPL)
 *
 * Known bugs: none
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include "sysmp.h"
#include "pset-utils.h"

char *progname = "";

int pset_create_all(char *);
int create_all_do_sysmp(unsigned long, unsigned long); 

void version(void)
{
	fprintf(stderr,"%s - version %s by Tim Hockin\n", progname, PSET_UTILS_VER);
}

void usage(void)
{
	version();
	fprintf(stderr, 
	"Usage: %s [-C conf_file] [-a]\n", progname); 
}

/* **********************************************************************
 * Main()
 * *********************************************************************/
int main(int argc, char *argv[])
{
	int retval = 0;
	char fn = 0;
	char *filename = PSETTAB;

	progname = argv[0];
	argc--; argv++;

	if (argc < 1 || argv[0][0] != '-') {
		usage();
		return -1;
	}
	
	while (argc) {
		switch(argv[0][1]) {
			case 'V':	
				if (fn) {
					usage();
					return -1;
				}
				fn = 'V';
				argc--; argv++;
				break;
			case 'a':
				if (fn) {
					usage();
					return -1;
				}
				fn = 'a';
				argc--; argv++;
				break;
			case 'C':
				if (argc < 2) {
					usage();
					return -1;
				}
				filename = argv[1];
				argc-=2; argv+=2;
				break;
			default: 	
				usage();
				return -1;
				break;
		}
	}
	switch (fn) {
		case 'a':
			/* create all psets in the conf file */
			retval = pset_create_all(filename);
			break;
		case 'V':	
			version();
			retval = 0;
			break;
		default:
			usage();
			retval = -1;
			break;
	}
			
	return retval;
}


#define DEBUG_FSM 0
#if DEBUG_FSM
#  define PRINTD(fmt, args...)	printf(fmt, ##args)
#else
#  define PRINTD(fmt, args...)
#endif
/* 
 * psettab entries are of the following form:
 * pset_name 	99	1,2,3,4 
 */
int pset_create_all(char *filename)
{
	int state = 0;
	int name_idx = 0;
	char num_buf[32];
	int num_idx = 0;
	int lineno = 1;
	unsigned long num = 0;
	unsigned long cpus = 0;
	char c;
	FILE *fp;
	enum states {
		STATE_BOL = 0,
		STATE_READ_NAME,
		STATE_POST_NAME,
		STATE_READ_NUM,
		STATE_POST_NUM,
		STATE_READ_CPULIST,
		STATE_POST_CPULIST,
		STATE_COMMENT,
		STATE_ERROR
	};

	fp = fopen(filename, "r");
	if (!fp) {
		perror(filename);
		return -1;
	}
	
	while ((c = fgetc(fp)) != EOF || state == STATE_ERROR) {
		switch (state) {
		case STATE_BOL:
			/* snarf up whitespace */
			PRINTD("BOL");
			num_idx = num = name_idx = cpus = 0;
			if (isspace(c)) {
				if (c == '\n') {
					lineno++;
				}
				state = state;
			} else if (c == '#') {
				state = STATE_COMMENT;
			} else if (isalpha(c) || c == '_') {
				name_idx++;
				state = STATE_READ_NAME;
			} else {
				state = STATE_ERROR;
			}
			break;
		case STATE_READ_NAME:
			/* step past the name of the pset */
			PRINTD("READ_NAME");
			if ((isalnum(c) || c == '_') && name_idx < 255) {
				name_idx++;
				state = state;
			} else if (isblank(c)) {
				state = STATE_POST_NAME;
			} else {
				state = STATE_ERROR;
			}
			break;
		case STATE_POST_NAME:
			/* eat space before a pset number */
			PRINTD("POST_NAME");
			if (isblank(c)) {
				state = state;
			} else if (isdigit(c)) {
				num_buf[num_idx++] = c;
				state = STATE_READ_NUM;
			} else {
				state = STATE_ERROR;
			}
			break;
		case STATE_READ_NUM:
			/* read a number */
			PRINTD("READ_NUM");
			if (isdigit(c) && num_idx < 31) {
				num_buf[num_idx++] = c;
			} else if (toupper(c) == 'X' && num_idx == 1) {
				num_buf[num_idx++] = 'x';
			} else if (isblank(c)) {
				num_buf[num_idx] = '\0';
				num = strtoul(num_buf, NULL, 0);
				PRINTD(" (num = %ld)", num);
				state = STATE_POST_NUM;
			} else {
				state = STATE_ERROR;
			}
			break;
		case STATE_POST_NUM:
			/* eat space before the CPU list */
			PRINTD("POST_NUM");
			if (isblank(c)) {
				state = state;
			} else if (isdigit(c)) {
				num_idx = 0;
				num_buf[num_idx++] = c;
				state = STATE_READ_CPULIST;
			} else {
				state = STATE_ERROR;
			}
			break;
		case STATE_READ_CPULIST:
			/* read the cpu list */
			PRINTD("READ_CPULIST");
			/* any of these terminate this cpu */
			if (isblank(c) || (c == '\n') || (c == ',')) {
				unsigned long cpu;
				num_buf[num_idx] = '\0';
				num_idx = 0;
				cpu = strtoul(num_buf, NULL, 0);
				PRINTD(" (cpu = %ld)", cpu);
				if (cpu > MAX_CPUS) {
					state = STATE_ERROR;
					break;
				}
				cpus |= 1 << cpu;
				PRINTD(" (cpus = 0x%08lx)", cpus);
			}
			if (isblank(c)) {
				/* we already handled this case */
				state = STATE_POST_CPULIST;
			} else if (c == '\n') {
				/* we already handled this case */
				lineno++;
				create_all_do_sysmp(num, cpus);
				state = STATE_BOL;
			} else if (c == ',') {
				/* we already handled this case */
				state = state;
			} else if (isdigit(c)) {
				num_buf[num_idx++] = c;
				num_buf[num_idx] = '\0';
				state = state;
			} else {
				state = STATE_ERROR;
			}
			break;
		case STATE_POST_CPULIST:
			/* eat until EOL */
			PRINTD("POST_CPULIST");
			if (isblank(c)) {
				state = state;
			} else if (c == '\n') {
				lineno++;
				create_all_do_sysmp(num, cpus);
				state = STATE_BOL;
			} else {
				state = STATE_ERROR;
			}
			break;
		case STATE_COMMENT:
			if (c == '\n') {
				lineno++;
				state = STATE_BOL;
			} else {
				state = state;
			}
			break;
		case STATE_ERROR:
			fclose(fp);
			fprintf(stderr, "%s: parse error on line %d\n", 
				PSETTAB, lineno);
			return -1;
			break;
		default:
			state = STATE_ERROR;
			break;
		}
		PRINTD(" %c\n", c); 
	}
	return 0;
}

/* helper func for pset_create_all - actually do the sysmp() call */
int create_all_do_sysmp(unsigned long num, unsigned long cpus) 
{
	int r;

	r = sysmp(MP_PSET, MPPS_CREATE, (int)num, cpus);
	if (r < 0) {
		fprintf(stderr, "Pset %ld: %s\n", num, strerror(errno));
	}

	return r;
}
