/*
 * runon
 * Run a process on a specified CPU
 * 
 * copyright 1998-1999 Tim Hockin
 * Released under the GNU Public License (GPL)
 *
 * Known bugs: none yet
 * Missing features: not that I am aware of
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include "sysmp.h"

void version()
{
	fprintf(stderr,"runon - version %s by Tim Hockin\n",PSET_UTILS_VER);
}

void usage()
{
	version();
	fprintf(stderr, 
	"Usage: runon -v | <pset_id | name> <command> [ <arguments> ... ]\n"); 
}

/* ****************************************************************
 * main()
 * 
 * At least two parameters :
 * 
 * argv[1]: pset to run command on
 * argv[2]: command to run
 * argv[3]  andd on are arguments to pass to exec()
 * 
 * ****************************************************************/
/* FIXME - make it handle if a pset name is specified */
int main(int argc, char *argv[])
{
	int pset_id;
	char *cmd;
  
	if (argc > 1 && !strncmp(argv[1], "-v", 2)) {
		version();
		return 0;
	} else if (argc < 3) {
		usage();
		return -1;
	}

	if (isdigit(argv[1][0])) {
		pset_id = atoi(argv[1]);
	} else {
		/* look it up in the psettab file */
		pset_id = get_pset_by_name(argv[1]);
  	}
  
	if (pset_id < 0) {
		fprintf(stderr, "pset %d is not a valid pset id or name\n", pset_id);
		return -1;
	}
  
	/* run on the specified CPU */
	if (sysmp(MP_MUSTRUN, pset_id) < 0) {
		if (errno == EINVAL) {
			fprintf(stderr, "pset %d does not exist\n", pset_id);
		} else {
			perror("sysmp(MUSTRUN)");
		}
		return -1;
	}
  
	cmd = argv[2];
	argv += 2;
  
	/* load the target */
	execvp(cmd, argv);
  
	/* if we get here, we didn't exec... */
	perror(argv[0]);

	return 1;
}
